#
# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import operator

from sys import version_info


PY2 = version_info.major == 2
PY3 = not PY2

if PY2:
    from functools32 import lru_cache

    reduce = reduce  # noqa

    def qualname(obj):
        """
        Returns the qualified name of `obj`.
        """
        return '.'.join(obj.__module__, obj.__name__)

    def items(dict_):
        return dict_.iteritems()

    def values(dict_):
        return dict_.itervalues()

else:
    from functools import lru_cache, reduce

    def qualname(obj):
        """
        Returns the qualified name of `obj`.
        """
        return obj.__qualname__

    def items(dict_):
        return dict_.items()

    def values(dict_):
        return dict_.values()


def _apply(a, b):
    return b(a)


def _composed_doc(fs):
    """
    Generate a docstring for the composition of fs.
    """
    if not fs:
        # Argument name for the docstring.
        return 'n'

    return '{f}({g})'.format(f=fs[0].__name__, g=_composed_doc(fs[1:]))


def compose(*fs):
    """
    Compose functions together in order:

    compose(f, g, h) = lambda n: f(g(h(n)))
    """
    # Pull the iterator out into a tuple so we can call `composed`
    # more than once.
    rs = tuple(reversed(fs))

    def composed(n):
        return reduce(lambda a, b: b(a), rs, n)

    # Attempt to make the function look pretty with
    # a fresh docstring and name.
    try:
        composed.__doc__ = 'lambda n: ' + _composed_doc(fs)
    except AttributeError:
        # One of our callables does not have a `__name__`, whatever.
        pass
    else:
        # We already know that for all `f` in `fs`, there exists `f.__name__`
        composed.__name__ = '_of_'.join(f.__name__ for f in fs)

    return composed


_nlname = '_NonLocal__nl'


class _MagicExpansionMeta(type):
    """
    A metaclass for expanding the @reflected and @inplace decorators.
    """
    def __new__(mcls, name, bases, dict_):
        for v in values(dict(dict_)):
            aliases = getattr(v, '_aliases', ())
            for alias in aliases:
                dict_[alias] = v

        return super(_MagicExpansionMeta, mcls).__new__(
            mcls, name, bases, dict_,
        )


def _alias(f, prefix):
    name = f.__name__
    if not name.startswith('__') and name.endswith('__'):
        raise ValueError('%s must be a dunder method' % name)

    name = name[2:-2]
    aliases = getattr(f, '_aliases', set())
    aliases.add('__%s%s__' % (prefix, name))
    f._aliases = aliases


def _reflected(f):
    """
    Aliases this magic with the reflected version.
    """
    _alias(f, 'r')
    return f


def _inplace(f):
    """
    Aliases this magic with the inplace version.
    """
    _alias(f, 'i')
    return f


class NonLocal(object):
    """
    NonLocal value for testing, backwards compat for py2.
    """
    def __init__(self, nl):
        object.__setattr__(self, _nlname, nl)

    def __getattribute__(self, attr):
        if attr == _nlname:
            return object.__getattribute__(self, attr)

        return getattr(self.__nl, attr)

    def __setattr__(self, attr, val):
        return setattr(self.__nl, attr, val)

    def __eq__(self, other):
        return self.__nl == other

    def __ne__(self, other):
        return self.__nl != other

    def __lt__(self, other):
        return self.__nl < other

    def __gt__(self, other):
        return self.__nl > other

    def __le__(self, other):
        return self.__nl <= other

    def __ge__(self, other):
        return self.__nl >= other

    def __pos__(self):
        return +self.__nl

    def __neg__(self):
        return -self.__nl

    def __abs__(self):
        return abs(self.__nl)

    def __invert__(self):
        return ~self.__nl

    def __round__(self):
        return round(self.__nl)

    def __floor__(self):
        return math.floor(self.__nl)

    def __ceil__(self):
        return math.ceil(self.__nl)

    def __trunc__(self):
        return math.truc(self.__nl)

    @_reflected
    @_inplace
    def __add__(self, other):
        return self.__nl + other

    @_reflected
    @_inplace
    def __sub__(self, other):
        return self.__nl - other

    @_reflected
    @_inplace
    def __mul__(self, other):
        return self.__nl * other

    @_reflected
    @_inplace
    def __floordiv__(self, other):
        return self.__nl // other

    @_reflected
    @_inplace
    def __div__(self, other):
        return self.__nl / other

    @_reflected
    @_inplace
    def __truediv__(self, other):
        return operator.truediv(self.__nl, other)

    @_reflected
    @_inplace
    def __mod__(self, other):
        return self.__nl % other

    @_reflected
    @_inplace
    def __divmod__(self, other):
        return divmod(self.__nl, other)

    @_reflected
    @_inplace
    def __pow__(self, other):
        return self.__nl ** other

    @_reflected
    @_inplace
    def __lshift__(self, other):
        return self.__nl << other

    @_reflected
    @_inplace
    def __rshift__(self, other):
        return self.__nl >> other

    @_reflected
    @_inplace
    def __and__(self, other):
        return self.__nl & other

    @_reflected
    @_inplace
    def __or__(self, other):
        return self.__nl | other

    @_reflected
    @_inplace
    def __xor__(self, other):
        return self.__nl ^ other

    def __int__(self):
        return int(self.__nl)

    def __float__(self):
        return float(self.__nl)

    def __complex__(self):
        return complex(self.__nl)

    def __oct__(self):
        return oct(self.__nl)

    def __hex__(self):
        return hex(self.__nl)

    def __index__(self):
        return self.__nl.__index__()

    def __str__(self):
        return str(self.__nl)

    def __bytes__(self):
        return bytes(self.__nl)

    def __repr__(self):
        return repr(self.__nl)

    def __format__(self, formatstr):
        return self.__nl.__format__(formatstr)

    def __hash__(self):
        return hash(self.__nl)

    def __bool__(self):
        return bool(self.__nl)

    def __dir__(self):
        return dir(self.__nl)

    def __delattr__(self, name):
        return delattr(self.__nl, name)

    def __len__(self):
        return len(self.__nl)

    def __getitem__(self, key):
        return self.__nl[key]

    def __setitem__(self, key, value):
        self.__nl[key] = value

    def __delitem__(self, key):
        del self.__nl[key]

    def __iter__(self):
        return iter(self.__nl)

    def __reversed__(self):
        return reversed(self.__nl)

    def __contains__(self, item):
        return item in self.__nl

    def __missing__(self, key):
        return self.__nl.__missing__[key]

    def __instancecheck__(self, instance):
        return isinstance(self.__nl, instance)

    def __call__(self, *args, **kwargs):
        return self.__nl(*args, **kwargs)

    def __enter__(self):
        return self.__nl.__enter__()

    def __exit__(self, exc_type, exc_value, exc_tb):
        return self.__nl.__exit__(exc_type, exc_value, exc_tb)

    def __get__(self, instance, owner):
        return self.__nl.__get__(instance, owner)

    def __set__(self, instance, value):
        return self.__nl.__get__(instance, value)

    def __delete__(self, instance):
        return self.__nl.__delete__(instance)

    def __prepare__(self, name, bases):
        return self.__nl.__prepare__(name, bases)

    def __next__(self):
        return next(self.__nl)

    # PY2 support:
    if PY2:
        __nonzero__ = __bool__
        # I can probably leave this; however, this is not the py2 convention.
        del __bool__

        def __coerce__(self, other):
            return self.__nl.__coerce__(other)

    @staticmethod
    def reassign(nl, val):
        object.__setattr__(nl, _nlname, val)


__all__ = [
    'NonLocal',
    'PY2',
    'PY3',
    'compose',
    'items',
    'lru_cache',
    'qualname',
    'reduce',
    'values',
]



#
# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import wraps
from textwrap import dedent

from metautils.box import methodbox
from metautils.compat import compose, items, lru_cache


class TemplateBase(object):
    """
    A marker for `Template` types.
    """
    pass


class _TemplateMeta(type):
    """
    Constructs `ClassTemplate` objects from type specs.
    This is meta.
    """
    def __new__(mcls,
                name,
                bases,
                dict_,
                preprocess=None,
                decorators=(),
                cachesize=None):

        template_param = bases[0]
        if not isinstance(template_param, _TemplateMeta):
            raise TypeError('class template must subclass T')

        bases = bases[1:]

        if issubclass(template_param, _TWithArgs):
            # The arguments were passed through the template param,
            # forward these to `__new__` to emulate python 3 class arguments.
            return mcls.__new__(
                mcls,
                name,
                (T,) + bases,  # Kill the infinite recursion.
                dict_,
                **template_param._kwargs
            )

        preprocess = preprocess or (lambda *a: a)
        decorators = tuple(decorators)

        class Template(TemplateBase):
            """
            A callable that takes a base metaclass and returns a new metaclass
            that is a composition of this metaclass template and a base
            metaclass.
            """
            __slots__ = ()

            def __call__(self, base=type, adjust_name=True):
                """
                Constructs a new metaclass that is the composition of this
                metaclass template and a base metaclass.

                If `adjust_name` is truthy, the name of the base class will be
                prepended with the name of the new class.
                """
                dict_cpy = dict_.copy()  # We could potentially mutate this.
                inner_bases = (base,) + bases

                # This function allows us to conditionally modify the class
                # name, bases, or dict after we have recieved the base
                # class. Think of this like a second meta layer.
                name_pp, inner_bases, dict_cpy = preprocess(
                    name, inner_bases, dict_cpy
                )

                inner_base = inner_bases[0]

                for k, v in items(dict_cpy):
                    if isinstance(v, templated):
                        # The method needs the base class in a closure, so we
                        # construct one here.
                        def close_by_value(unboxed=v.unboxed):
                            # This is the dumbest python scoping issue.
                            @wraps(unboxed)
                            def wrapper(*args, **kwargs):
                                return unboxed(*args, T_=inner_base, **kwargs)

                            return wrapper

                        dict_cpy[k] = close_by_value()

                if adjust_name:
                    # We want to have the base's name prepended to ours.
                    name_pp = base.__name__ + name_pp

                tp = compose(*decorators)(type(name_pp, inner_bases, dict_cpy))
                # This is Python 3 specific, but there is no need to make
                # a check as this will not fail in Python 2, it is just
                # not used.
                tp.__qualname__ = name_pp
                return tp

            if cachesize is None or cachesize >= 0:
                __call__ = lru_cache(cachesize)(__call__)

            def __repr__(self):
                return '<{cls}: {name} at 0x{id_}>'.format(
                    cls=type(self).__name__,
                    name=name,
                    id_=hex(id(self)),
                )

            __str__ = __repr__

        return Template()


class _TWithArgs(object):
    """
    Marker to indicate that this is a template argument that is holding the
    arguments.
    """
    pass


def T_new(cls, **kwargs):
    """
    `__new__` for `T`. This allows us to pass argument to the metaclass
    in Python 2.
    """
    # We are using `type.__new__` so that we get a concrete class, not
    # a template.
    return type.__new__(_TemplateMeta, 'TWithArgs', (_TWithArgs,), {
        '__doc__': 'A template parameter with arguments applied.',
        '__slots__': (),
        '_kwargs': kwargs,
    })


# The template parameter. This uses `type.__new__` because we need a concrete
# class here, not a template.
T = type.__new__(_TemplateMeta, 'T', (object,), {
    '__doc__': dedent(
        """
        A template parameter similar to c++ class templates.

        Class templates are defined by 'subclassing' `T`.
        For example:

        >>> class MyTemplate(T):
        ...    @templated
        ...    def method(self, T_):
        ...        print(T_, 'is the template argument')

        To construct instances of your templated class, call the
        template with the class you want to subclass.

        >>> NewClass = MyTemplate(MyBaseClass)
        >>> NewClass.__base__ == MyBaseClass
        True
        >>> NewClass.__mro__ = (MyBaseClass,) + MyBaseClass.__mro__
        True

        The `templated` decorator will pass the templated argument to the
        function implicitly allowing us to close over the base class.

        >>> NewClass().method()
        'MyBaseClass is the template argument'

        Additional arguments can be passed to the template as keyword
        arguments in the class statement, or as arguments to `T`.

        >>> class Python3Style(T, decorators=(mydecorator,)):
        ...     pass

        or in Python 2 without class arguments:

        >>> class Python2Style(T(decorators=(mydecorator,))):
        ...     pass

        Python 2 style will work in Python 2 or 3, so it should be used
        for cross compatible code; however, the Python 3 style looks
        much better and should be used if no python 2 compatibility is
        needed. In Python 3, passing arguments through the template
        argument will override the arguments passed to the class by keyword.

        The arguments may include:

        preprocess: A function that takes the name, bases, and dict_
          of the class before it is generated and returns a new name,
          bases, and dict_. This allows you to change any of these
          class parameters before the class is constructed but after
          the template argument has been passed. The template argument
          will be `bases[0]`.

        decorators: An iterable of class decorators to apply to the
          newly constructed class object.

        cachesize: Because templates are normally used to construct
          classes dynamically, we frequently will pass the same base
          classes in multiple places. To make this more efficient,
          we can `lru_cache` the class factory. This is the argument
          that will be the `lru_cache`'s cachesize. If this is less
          than 0, no cache will be used. If this is `None`, then there
          will be no upper bounds on the cache.
        """,
    ),
    '__new__': T_new,
    '__slots__': (),
})


class templated(methodbox):
    """
    Marker to indicate that the method should be passed `T` under the
    name: `T_`
    """
    pass



#!/usr/bin/env python
from distutils.core import setup
import sys

long_description = ''

if 'upload' in sys.argv:
    with open('README.rst') as f:
        long_description = f.read()

version_info = sys.version_info

setup(
    name='metautils',
    version='0.1.1',
    description='Utilities for working with metaclasses.',
    author='Quantopian Inc.',
    author_email='opensource@quantopian.com',
    packages=[
        'metautils',
    ],
    long_description=long_description,
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python',
        'Topic :: Utilities',
    ],
    url='https://github.com/quantopian/metautils',
    install_requires=(
        'functools32',
    ) if version_info.major == 2 and version_info == 7 else (),
)
